from __future__ import annotations

import logging
import os
from collections import defaultdict
from typing import Any

import numpy as np
import pandas as pd

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)


def read_and_sort_triplets(triplets_path: str) -> pd.DataFrame:
    """Read triplets from a file and sort by the 'Score' column."""
    triplets = pd.read_csv(triplets_path, sep='\t')
    return triplets.sort_values(by='Score', ascending=False)


def generate_final_ranking(triplets: pd.DataFrame, label_column: str) -> list[list[Any]]:
    """Generate final ranking based on the label column."""
    final_ranking = []
    for _, row in triplets.iterrows():
        feature_a, feature_b = row['FeatureA'], row['FeatureB']
        score = row['Score']
        if label_column == feature_a.split('-')[0]:
            final_ranking.append([feature_b, score])
        elif label_column == feature_b.split('-')[0]:
            final_ranking.append([feature_a, score])
    return final_ranking


def create_final_dataframe(final_ranking: list[list[Any]], heuristic: str) -> pd.DataFrame:
    """Create a final DataFrame and normalize if necessary."""
    final_df = pd.DataFrame(final_ranking, columns=['Feature', f'Score {heuristic}'])
    final_df = (
        final_df.groupby('Feature')
        .median()
        .reset_index()
        .sort_values(by=f'Score {heuristic}', ascending=False)
    )

    if 'MI' in heuristic:
        min_score = final_df[f'Score {heuristic}'].min()
        max_score = final_df[f'Score {heuristic}'].max()
        final_df[f'Score {heuristic}'] = (final_df[f'Score {heuristic}'] - min_score) / (max_score - min_score)

    return final_df


def store_summary_files(final_df: pd.DataFrame, output_folder: str, heuristic: str, tldr: bool) -> None:
    """Store the summary files and optionally print the head of the DataFrame."""
    logging.info(f'Storing summary files to {output_folder}')
    pd.set_option('display.max_rows', None, 'display.max_columns', None)

    singles_path = os.path.join(output_folder, 'feature_singles.tsv')
    final_df.to_csv(singles_path, sep='\t', index=False)

    if tldr:
        print(final_df.head(20))


def handle_interaction_order(final_df: pd.DataFrame, output_folder: str, heuristic: str, interaction_order: int) -> None:
    """Handle the interaction order if it is greater than 1."""
    if interaction_order > 1:
        feature_store = defaultdict(list)
        for _, row in final_df.iterrows():
            fname = row['Feature']
            score = row[f'Score {heuristic}']
            if 'AND' in fname:
                for el in fname.split('-')[0].split(' AND '):
                    feature_store[el].append(score)

        final_aggregate_df = pd.DataFrame([
            {
                'Feature': k,
                f'Combined score (order: {interaction_order}, {heuristic})': np.median(v),
            }
            for k, v in feature_store.items()
        ])
        final_aggregate_df.to_csv(
            os.path.join(output_folder, 'feature_singles_aggregated.tsv'), sep='\t', index=False,
        )


def filter_transformers_only(final_df: pd.DataFrame, output_folder: str) -> None:
    """Filter the DataFrame to include only transformer features and store the result."""
    transformers_only_path = os.path.join(output_folder, 'feature_singles_transformers_only_imp.tsv')
    final_df[final_df['Feature'].str.contains('_tr_')].to_csv(transformers_only_path, sep='\t', index=False)


def outrank_task_result_summary(args) -> None:
    """Main function to generate a summary of outrank task results."""
    triplets_path = os.path.join(args.output_folder, 'pairwise_ranks.tsv')
    triplets = read_and_sort_triplets(triplets_path)

    final_ranking = generate_final_ranking(triplets, args.label_column)
    final_df = create_final_dataframe(final_ranking, args.heuristic)

    store_summary_files(final_df, args.output_folder, args.heuristic, args.tldr)
    handle_interaction_order(final_df, args.output_folder, args.heuristic, args.interaction_order)
    filter_transformers_only(final_df, args.output_folder)



# A collection of feature transformers that can be considered
from __future__ import annotations

import logging
from typing import Any

import numpy as np
import pandas as pd

import outrank.feature_transformations.feature_transformer_vault as transformer_vault
from outrank.core_utils import internal_hash

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)


class FeatureTransformerNoise:
    def __init__(self):
        self.noise_preset = 'default'

    def construct_new_features(self, dataframe: pd.DataFrame, label_column=None):
        """Generate a few standard noise distributions"""

        new_columns = dict()
        if self.noise_preset == 'default':
            new_columns['CONTROL-constant0'] = np.array([0] * dataframe.shape[0])
            new_columns['CONTROL-gaussian'] = np.random.normal(
                size=dataframe.shape[0],
            )
            new_columns['CONTROL-uniform'] = np.random.random(
                dataframe.shape[0],
            )
            new_columns['CONTROL-random-binary'] = np.random.randint(
                0, 2, dataframe.shape[0],
            )
            new_columns['CONTROL-random-card100'] = np.random.randint(
                0, 1 + 1 * 10**2, dataframe.shape[0],
            )
            new_columns['CONTROL-random-card2k'] = np.random.randint(
                0, 1 + 2 * 10**3, dataframe.shape[0],
            )
            new_columns['CONTROL-random-card10k'] = np.random.randint(
                0, 1 + 10 * 10**3, dataframe.shape[0],
            )
            new_columns['CONTROL-random-card50k'] = np.random.randint(
                0, 1 + 50 * 10**3, dataframe.shape[0],
            )
            new_columns['CONTROL-int-sequence'] = np.arange(
                0, dataframe.shape[0], 1.0,
            )

            if label_column not in dataframe.columns:
                logging.warn(
                    'Could not find target feature in your data set - please inspect the columns if doing targeted ranking!',
                )
            else:
                new_columns['CONTROL-target'] = dataframe[label_column]

            new_columns['CONTROL-volume'] = np.array([
                internal_hash(str(x)) for _, x in dataframe.iterrows()
            ])
        else:
            # Not relevant yet; will be if this is useful.
            pass

        if len(new_columns) > 0:
            tmp_df = pd.DataFrame(new_columns)
            dataframe = pd.concat([dataframe, tmp_df], axis=1)
            del tmp_df

        return dataframe


class FeatureTransformerGeneric:
    def __init__(self, numeric_column_names: set[str], preset: str = 'default'):
        for transformer_namespace in preset.split(','):
            self.transformer_collection: dict[str, str] = dict()
            transformer_subspace = transformer_vault._tr_global_namespace.get(
                transformer_namespace, None,
            )
            if transformer_subspace:
                self.transformer_collection = {
                    **self.transformer_collection,
                    **transformer_subspace,
                }

            if len(self.transformer_collection) == 0:
                raise NotImplementedError(
                    'Please, specify valid transformer namespaces (e.g., default, minimal etc.)',
                )

        self.numeric_column_names = set(numeric_column_names)
        self.constructed_feature_names: set[str] = set()

        # If 80% of values are the same, don't consider a transformation
        self.max_maj_support = 0.80

        # If more than 75% of vals are missing, don't consider a transformation
        self.nan_prop_support = 0.75

    def get_vals(self, tmp_df: pd.DataFrame, col_name: str) -> Any:
        cvals = tmp_df[col_name].values.tolist()
        cvals = [str(x).replace('"', '') for x in cvals]
        cvals = [0.0 if len(x) == 0 else float(x) for x in cvals]

        return np.array(cvals)

    def construct_baseline_features(self, dataframe: Any) -> pd.DataFrame:
        fvals = []
        for enx, row in dataframe.iterrows():
            missing_prop = np.round(
                row.values.tolist().count('') / dataframe.shape[1], 1,
            )
            fvals.append(missing_prop)

        dataframe['BASELINE-MISSING-PROPORTION'] = fvals
        dataframe['BASELINE-DUMMY'] = 0

        return dataframe

    def construct_new_features(self, dataframe: Any) -> pd.DataFrame:
        new_numeric = set()
        logging.info(
            f'Considering {len(self.transformer_collection)} transformations for {len(self.numeric_column_names)} features ({len(self.transformer_collection) * len(self.numeric_column_names)} new features will be considered).',
        )

        invalid_transforms = 0
        new_columns = dict()
        for numeric_column in self.numeric_column_names:
            X = self.get_vals(dataframe, numeric_column)

            if len(X) == 0:
                raise AssertionError(
                    f"Could not retrieve the colomn {numeric_column}'s values. Please check the data.",
                )

            for k, v in self.transformer_collection.items():
                feature_name = f'{numeric_column}{k}'
                transformed_array = eval(v).astype(str)
                u, c = np.unique(transformed_array, return_counts=True)
                nan_prop = np.count_nonzero(transformed_array == 'nan') / len(
                    transformed_array,
                )
                cfreq = np.divide(np.max(c), np.sum(c))
                if (
                    len(u) > 1
                    and cfreq < self.max_maj_support
                    and nan_prop < self.nan_prop_support
                ):
                    new_columns[feature_name] = transformed_array
                    new_numeric.add(feature_name)

                else:
                    invalid_transforms += 1

        if len(new_columns) > 0:
            tmp_df = pd.DataFrame(new_columns)
            dataframe = pd.concat([dataframe, tmp_df], axis=1)
            del tmp_df

        logging.info(
            f'{invalid_transforms} invalid transformations were skipped.',
        )
        self.numeric_column_names = self.numeric_column_names
        self.constructed_feature_names = new_numeric
        return dataframe



from __future__ import annotations

import gzip
import os
from collections import Counter
from collections import defaultdict
from typing import Any

import numpy as np
import pandas as pd
import tqdm

from outrank.core_utils import generic_line_parser
from outrank.core_utils import get_dataset_info
from outrank.core_utils import get_num_of_instances

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

def shannon_entropy(string: str) -> float:
    counts = Counter(string)
    frequencies = (i / len(string) for i in counts.values())
    return -sum(f * np.log2(f) for f in frequencies)

def compute_average_entropy(line: list[str]) -> float:
    return sum(shannon_entropy(field) for field in line)

def score_line(line: list[str]) -> dict[str, float]:
    total_fields = len(line)
    nan_prop = line.count('') / total_fields
    empty_dict_prop = line.count('{}') / total_fields
    all_empty_prop = (line.count('{}') + line.count('')) / total_fields
    all_zero_prop = line.count('0') / total_fields

    out_struct = {
        'empty_string_prop': nan_prop,
        'empty_dict': empty_dict_prop,
        'all_empty': all_empty_prop,
        'all_zero': all_zero_prop,
        'row_entropy': compute_average_entropy(line),
    }

    for j in [30, 60, 100, 200, 300]:
        out_struct[f'all_more_{j}_chars'] = sum(len(x) > j for x in line) / total_fields

    return out_struct

def outrank_task_rank_instances(args: Any) -> None:
    dataset_info = get_dataset_info(args)
    data_path = dataset_info.data_path
    data_encoding = 'utf-8'
    delimiter = '\t'

    total_lines = get_num_of_instances(data_path) - 1
    local_pbar = tqdm.tqdm(total=total_lines, position=0, disable=args.disable_tqdm == 'True')
    local_pbar.set_description('Starting ranking computation')

    _, file_extension = os.path.splitext(data_path)
    file_stream = gzip.open(data_path, 'rt', encoding=data_encoding) if file_extension == '.gz' else open(data_path, encoding=data_encoding)

    line_counter = 0
    out_scores_lab = defaultdict(list)

    for line in file_stream:
        line_counter += 1
        local_pbar.update(1)

        parsed_line = generic_line_parser(
            line,
            delimiter,
            args,
            dataset_info.fw_map,
            dataset_info.column_names,
        )

        if line_counter > 100_000:
            break
        out_scores_lab[line[0]].append(score_line(parsed_line))

    file_stream.close()

    os.makedirs(args.output_folder, exist_ok=True)
    for label, out_scores in out_scores_lab.items():
        out_df = pd.DataFrame(out_scores)
        for col in out_df.columns:
            sorted_vals = out_df[col].sort_values()
            plt.figure(figsize=(5, 5), dpi=300)
            plt.title(f'{col} label: {label}')
            plt.hist(
                x=sorted_vals * 100,
                color='black',
                density=True,
                bins=100,
            )
            plt.xlabel('Proportion of namespaces (%)' if 'entropy' not in col else 'Row entropy')
            plt.ylabel('Density')
            plt.tight_layout()
            fname = f'distPlot{col}_{label}.pdf'
            plt.savefig(os.path.join(args.output_folder, fname), dpi=300)
            plt.cla()
            plt.clf()

    local_pbar.close()



from __future__ import annotations

import logging
import operator
import traceback
from typing import Any

import numpy as np
import pandas as pd
from scipy.stats import pearsonr
from sklearn import random_projection
from sklearn.decomposition import TruncatedSVD
from sklearn.feature_selection import mutual_info_classif
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model import SGDClassifier
from sklearn.metrics import adjusted_mutual_info_score
from sklearn.model_selection import cross_val_score
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder
from sklearn.svm import SVC

from outrank.algorithms.feature_ranking import ranking_cov_alignment

logger = logging.getLogger('syn-logger')
logger.setLevel(logging.DEBUG)

NUM_FOLDS  = 2
SVD_DIMS = 2

try:
    from outrank.algorithms.feature_ranking import ranking_mi_numba
    numba_available = True
except ImportError:
    traceback.print_exc()
    numba_available = False

def sklearn_MI(vector_first: np.ndarray, vector_second: np.ndarray) -> float:
    return mutual_info_classif(
        vector_first.reshape(-1, 1), vector_second.reshape(-1), discrete_features=True,
    )[0]

def sklearn_surrogate(
    vector_first: np.ndarray, vector_second: np.ndarray,  surrogate_model: str,
) -> float:
    X = OneHotEncoder().fit_transform(vector_first)

    if '-SVD' in surrogate_model and X.shape[1] > 2:
        # yes this is not super correct due to embedding full data first, but it's much faster + seems to offer same results anyways.
        X = TruncatedSVD(n_components=min(SVD_DIMS, X.shape[1])).fit_transform(X)

    clf = initialize_classifier(surrogate_model, n_dim=min(X.shape[1], 1024))
    scores = cross_val_score(clf, X, vector_second, scoring='neg_log_loss', cv=NUM_FOLDS)
    return 1 + np.median(scores)

def numba_mi(vector_first: np.ndarray, vector_second: np.ndarray, heuristic: str, mi_stratified_sampling_ratio: float) -> float:
    cardinality_correction = heuristic == 'MI-numba-randomized'

    try:
        if vector_first.shape[1] == 1:
            vector_first = vector_first.reshape(-1)
        else:
            vector_first = np.apply_along_axis(lambda x: np.abs(np.max(x) - np.sum(x)), 1, vector_first).reshape(-1)
    except:
        logger.warning('Reshaping for MI computation in place - you are considering many-one mapping')

    return ranking_mi_numba.mutual_info_estimator_numba(
        vector_first.astype(np.int32),
        vector_second.astype(np.int32),
        approximation_factor=np.float32(mi_stratified_sampling_ratio),
        cardinality_correction=cardinality_correction,
    )

def sklearn_mi_adj(vector_first: np.ndarray, vector_second: np.ndarray) -> float:
    return adjusted_mutual_info_score(vector_first, vector_second)

def generate_data_for_ranking(combination: tuple[str, str], reference_model_features: list[str], args: Any, tmp_df: pd.DataFrame) -> tuple(np.ndarray, np.ndrray):
    feature_one, feature_two = combination

    if feature_one == args.label_column:
        feature_one = feature_two
        feature_two = args.label_column

    if args.reference_model_JSON:
        vector_first = tmp_df[list(reference_model_features) + [feature_one]].values
    else:
        vector_first = tmp_df[feature_one].values

    vector_second = tmp_df[feature_two].values
    return vector_first, vector_second


def conduct_feature_ranking(vector_first: np.ndarray, vector_second: np.ndarray, args: Any) -> float:

    heuristic = args.heuristic
    score = 0.0

    if heuristic == 'MI':
        score = sklearn_MI(vector_first, vector_second)

    elif heuristic in {'surrogate-SGD', 'surrogate-SVM', 'surrogate-SGD-RP', 'surrogate-SGD-SVD'}:
        score = sklearn_surrogate(vector_first, vector_second, heuristic)

    elif heuristic == 'max-value-coverage':
        score = ranking_cov_alignment.max_pair_coverage(vector_first, vector_second)

    elif heuristic == 'MI-numba-randomized':
        score = numba_mi(vector_first, vector_second, heuristic, args.mi_stratified_sampling_ratio)

    elif heuristic == 'AMI':
        score = sklearn_mi_adj(vector_first, vector_second)

    elif heuristic == 'correlation-Pearson':
        score = pearsonr(vector_first, vector_second)[0]

    elif heuristic == 'Constant':
        score = 0.0

    else:
        logger.warning(f'{heuristic} not defined!')
        score = 0.0

    return score

def get_importances_estimate_pairwise(combination: tuple[str, str], reference_model_features: list[str], args: Any, tmp_df: pd.DataFrame) -> tuple[str, str, float]:

    feature_one, feature_two = combination
    inputs_encoded, output_encoded = generate_data_for_ranking(combination, reference_model_features, args, tmp_df)

    ranking_score = conduct_feature_ranking(inputs_encoded, output_encoded, args)

    return feature_one, feature_two, ranking_score


def rank_features_3MR(
    relevance_dict: dict[str, float],
    redundancy_dict: dict[tuple[Any, Any], Any],
    relational_dict: dict[tuple[Any, Any], Any],
    strategy: str = 'median',
    alpha: float = 1.0,
    beta: float = 1.0,
) -> pd.DataFrame:
    all_features = set(relevance_dict.keys())
    most_important_feature = max(relevance_dict.items(), key=operator.itemgetter(1))[0]
    ranked_features = [most_important_feature]

    def calc_higher_order(feature: str, is_redundancy: bool = True) -> float:
        values = []
        for feat in ranked_features:
            interaction_tuple = (feat, feature)
            if is_redundancy:
                values.append(redundancy_dict.get(interaction_tuple, 0))
            else:
                values.append(relational_dict.get(interaction_tuple, 0))
        return np.median(values) if strategy == 'median' else (np.mean(values) if strategy == 'mean' else sum(values))

    while len(ranked_features) < len(all_features):
        top_importance = -np.inf
        most_important_feature = None

        for feat in all_features - set(ranked_features):
            feature_redundancy = calc_higher_order(feat)
            feature_relation = calc_higher_order(feat, False)
            feature_relevance = relevance_dict[feat]
            importance = feature_relevance - alpha * feature_redundancy + beta * feature_relation

            if importance > top_importance:
                top_importance = importance
                most_important_feature = feat

        ranked_features.append(most_important_feature)

    return pd.DataFrame({'Feature': ranked_features, '3MR_Ranking': range(1, len(ranked_features) + 1)})

def get_importances_estimate_nonmyopic(args: Any, tmp_df: pd.DataFrame):
    pass

def initialize_classifier(surrogate_model: str, n_dim: int) -> Any:

    if 'surrogate-LR' in surrogate_model:
        return LogisticRegression(max_iter=100000)

    elif 'surrogate-SVM' in surrogate_model:
        return SVC(gamma='auto', probability=True)

    elif 'surrogate-SGD-RP' in surrogate_model:
        clf = Pipeline([('proj', random_projection.SparseRandomProjection(n_components=n_dim)), ('reg', SGDClassifier(max_iter=100000, loss='log_loss'))])
        return clf

    elif 'surrogate-SGD' in surrogate_model:
        return SGDClassifier(max_iter=100000, loss='log_loss')

    else:
        logger.warning(f'The chosen surrogate model {surrogate_model} is not supported, falling back to surrogate-SGD')
        return SGDClassifier(max_iter=100000, loss='log_loss')
